/*
 * Copyright (C) 2006 Jonathan McDowell <noodles@earth.li>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include "delay.h"
#include "types.h"
#include "uart.h"

#define ADDR_COLUMN			1
#define ADDR_PAGE			2
#define ADDR_COLUMN_PAGE		3

#define AMS_DELTA_LATCH2_NAND_NCE	0x0004
#define AMS_DELTA_LATCH2_NAND_NRE	0x0008
#define AMS_DELTA_LATCH2_NAND_NWP	0x0010
#define AMS_DELTA_LATCH2_NAND_NWE	0x0020
#define AMS_DELTA_LATCH2_NAND_ALE	0x0040
#define AMS_DELTA_LATCH2_NAND_CLE	0x0080

#define MPUIO_DATA_INPUT_REG		0xFFFB5000
#define MPUIO_DATA_OUTPUT_REG		0xFFFB5004
#define MPUIO_DIR_CONTROL_REG		0xFFFB5008

#define NAND_CMD_READ0			0x00
#define NAND_CMD_READOOB		0x50
#define NAND_CMD_STATUS			0x70
#define NAND_CMD_READID			0x90
#define NAND_CMD_RESET			0xff

u16 ams_delta_latch2_val;

void ams_delta_latch2_write(u16 mask, u16 value)
{
	ams_delta_latch2_val &= ~mask;
	ams_delta_latch2_val |= value;

	*(volatile u16 *) 0x08000000 = ams_delta_latch2_val;
}

u8 ams_delta_nand_read(void)
{
	u8 val;

	ams_delta_latch2_write(AMS_DELTA_LATCH2_NAND_NRE |
			AMS_DELTA_LATCH2_NAND_NWE,
			AMS_DELTA_LATCH2_NAND_NWE);
	*(volatile u8 *) MPUIO_DIR_CONTROL_REG = ~0;
	udelay(1);
	val = *(volatile u8 *) MPUIO_DATA_INPUT_REG;
	ams_delta_latch2_write(AMS_DELTA_LATCH2_NAND_NRE,
				AMS_DELTA_LATCH2_NAND_NRE);

	return val;
}

void ams_delta_nand_write(u8 val)
{
/*	uart_puts("Writing: ");
	uart_putbyte(val);
	uart_puts("\r\n"); */

	*(volatile u8 *) MPUIO_DIR_CONTROL_REG = 0;
	*(volatile u8 *) MPUIO_DATA_OUTPUT_REG = val;
	ams_delta_latch2_write(AMS_DELTA_LATCH2_NAND_NWE |
				AMS_DELTA_LATCH2_NAND_NRE,
				AMS_DELTA_LATCH2_NAND_NRE);
	udelay(1);
	ams_delta_latch2_write(AMS_DELTA_LATCH2_NAND_NWE,
			AMS_DELTA_LATCH2_NAND_NWE);
}

void nand_wait_ready(void)
{
	/*
	 * FIXME NAND RB is on GPIO 12, but didn't work with u-boot?
	 */
	udelay(25);
}

void nand_address(int numbytes, unsigned long ofs)
{
	int i;

	/* Assert ALE */
	ams_delta_latch2_write(AMS_DELTA_LATCH2_NAND_ALE,
			AMS_DELTA_LATCH2_NAND_ALE);

	/* Send the address */
	if (numbytes == ADDR_COLUMN || numbytes == ADDR_COLUMN_PAGE)
		ams_delta_nand_write(ofs);

	ofs = ofs >> 9;
	if (numbytes == ADDR_PAGE || numbytes == ADDR_COLUMN_PAGE) {
		for (i = 0; i < 2; i++, ofs = ofs >> 8) {
			ams_delta_nand_write(ofs);
		}
	}

	/* Clear ALE */
	ams_delta_latch2_write(AMS_DELTA_LATCH2_NAND_ALE, 0);

	nand_wait_ready();
}

void nand_command(u8 command)
{
	u8 ret;

	/* Assert CLE */
	ams_delta_latch2_write(AMS_DELTA_LATCH2_NAND_CLE,
				AMS_DELTA_LATCH2_NAND_CLE);
	/* Write the command */
	ams_delta_nand_write(command);
	/* Lower CLE */
	ams_delta_latch2_write(AMS_DELTA_LATCH2_NAND_CLE, 0);

	if (command == NAND_CMD_RESET) {
		nand_command(NAND_CMD_STATUS);
		do {
			ret = ams_delta_nand_read();
		} while ((ret & 0x40) != 0x40);
	}

	nand_wait_ready();
}

int nand_ident(void)
{
	u8 mfr, id;

	/* Enable CE */
	ams_delta_latch2_write(AMS_DELTA_LATCH2_NAND_NCE, 0);
	/* Reset the chip */
	nand_command(NAND_CMD_RESET);
	/* Read the chip ID: #1 Send ReadID command */
	nand_command(NAND_CMD_READID);
	/* Read the chip ID: #2 Send address byte 0 */
	nand_address(ADDR_COLUMN, 0);

	udelay(10);
	/* Read the manufacturer and device id codes. */
	mfr = ams_delta_nand_read();
	id = ams_delta_nand_read();
	/* Disble CE */
	ams_delta_latch2_write(AMS_DELTA_LATCH2_NAND_NCE,
			AMS_DELTA_LATCH2_NAND_NCE);
	
	if (id == 0x75 && mfr == 0x98) {
		uart_puts("Found Tosh NAND.\r\n");
		return 1;
	}
	
	return 0;
}

void nand_readbuf(size_t len, u8 *buf)
{
	nand_command(NAND_CMD_READ0);

	while (len > 0) {
		*buf++ = ams_delta_nand_read();
		len--;
	}
}

int nand_read_oob(size_t ofs, size_t len, u8 *buf)
{
	int ret = 0;

	/* Enable CE */
	ams_delta_latch2_write(AMS_DELTA_LATCH2_NAND_NCE, 0);

	nand_command(NAND_CMD_READOOB);
	nand_address(ADDR_COLUMN_PAGE, ofs);

	nand_readbuf(len, buf);

	nand_wait_ready();

	/* Disable CE */
	ams_delta_latch2_write(AMS_DELTA_LATCH2_NAND_NCE, 0);

	return ret;
}

int nand_check_block(unsigned long pos)
{
	u8 oob_data;
	unsigned long page0 = pos & (-0x4000);
	unsigned long page1 = page0 + 512;
	
	nand_read_oob(page0 + 5, 1, &oob_data);
	if (oob_data != 0xff)
		return 1;
	nand_read_oob(page1 + 5, 1, &oob_data);
	if (oob_data != 0xff)
		return 1;

	return 0;
}

void nand_print_badblocks(void)
{
	int i;

	uart_puts("Checking NAND bad blocks:\r\n");

	
	for (i = 0; i < 32 * 1024 * 1024; i += 0x4000) {
		if ((i & 0xFFFFF) == 0) {
			uart_puts("At 0x");
			uart_putword(i);
			uart_puts("\r\n");
		}
		if (nand_check_block(i)) {
			uart_puts("\t");
			uart_putword(i);
			uart_puts("\r\n");
		}
	}
}

void nand_boot(void)
{
	int page;
	int start = 0x380000;
	int len;
	u8 *mem;
	void (*uboot)(void);

	mem = (void *) 0x11f00000;

	page = start >> 9;
	len = 0;

	uart_puts("Loading u-boot");
	while (len < (128 * 1024)) {
		if ((page % 32) == 0) {
			while (nand_check_block(page * 512)) {
				/*
				 * Block is bad; skip it and print a ! 
				 */
				uart_puts("!");
				page += 32;
			}
		}
		ams_delta_latch2_write(AMS_DELTA_LATCH2_NAND_NCE, 0);

		nand_command(NAND_CMD_READ0);

		nand_address(ADDR_COLUMN_PAGE, (page << 9));
		nand_readbuf(512, mem + len);


		ams_delta_latch2_write(AMS_DELTA_LATCH2_NAND_NCE, 
					AMS_DELTA_LATCH2_NAND_NCE);

		len += 512;
		page++;
		uart_puts(".");
	}

	uart_puts("\r\nStarting u-boot.\r\n\n");
	uboot = mem;
	uboot();
}

int nand_init(void)
{
	/*
	 * Disable CE, RE, WE.
	 */
	ams_delta_latch2_write(~0, 0);

	return nand_ident();
}
